# Neo Luminosity & Lightness Mask  Maker 8 & 16bit Version
# Ver. 0.1 reactive version in developping
# add tranceparency factor
# add luminosity mask
# add Nega mask (2022.5)
# add binded preview image (2022.8)
# (c)2021-2 Ohnishi, Yasuo

from ij import IJ, ImagePlus
from ij import WindowManager as WM
from ij.gui import GenericDialog, ImagePanel
from ij.io import DirectoryChooser, OpenDialog
from ij import plugin
from ij.plugin import ChannelSplitter, Duplicator
from ij.plugin import RGBStackMerge, RGBStackConverter
from ij.plugin.filter import ThresholdToSelection
from ij.process import ImageConverter, ImageProcessor
from ij.process import ByteProcessor
from loci.plugins import BF
from java.awt.event import AdjustmentListener, ItemListener, TextListener
import os
from org.python.core import codecs
codecs.setDefaultEncoding('utf-8')

#---Convert Value in 8, 16, 32bit mode
def convValue(Value, Mode):
#     Converting Luminance Value each in 8, 16, 32bit mode
        if Mode == "16-bit":
                Value = Value * 256
        elif Mode == "32-bit":
                Value = Value * 256 * 256
        else:
                Value = Value
        return Value

#--------------------------------------------
def luminosity_mask_maker(imp1, LTh, Mode, NegaInvert, Show, TrFac, chName):
# imp1: Grayscale image
# LTh: Zone lower Threshold
# NegaInvert: Nega invert True or False
# TrFac: Transparency Factor
# Close Windows
#    IJ.run(imp1, "Conversions...", "weighted")
    if Show == "Yes":
        imp1.show()
# Make Grayscale Mask Image Files 
    V1 = str(LTh) # for threshold expression in file name
# Target (Yellow) Mask Threshold
    V2 = str(convValue(abs(LTh), Mode)) # actual Yellow mask Threshold
    IJ.run(imp1, "Subtract...","value=" + V2)
    if NegaInvert == True: # for making inverted Negative mask
        IJ.run(imp1, "Divide...", "value=" + str(TrFac)) 
    else:
        IJ.run(imp1, "Multiply...", "value=" + str(TrFac)) 
            # making mask more transparent (x 4) < canceled
    if NegaInvert == True: # for making inverted Negative mask
        IJ.run(imp1, "Invert","")
        V1 = V1 + "_Nega"
        V = "_Nega_x "+ str(TrFac)
    else: # for making positive mask
        V1 = V1 + "_Posi"
        V = "_Posi_x" + str(TrFac)
    impGraymask = imp1
    if Show == "Yes":
        impGraymask.show()  # Show the subtracted image in window
    IJ.saveAsTiff(impGraymask, folder + filename + "_" + chName + "_Mask_Thr_".decode('utf-8')\
                        + V1.decode('utf-8') + "_x4x" + str(TrFac) + ". tif".decode('utf-8')) 
             # Save the Subtracted image x 4 x transparency factor
#   save grayscalee mask
    return impGraymask, V

#-------------------------------------------
def make_binary_mask(impZone, zoneLTh, zoneHTh, Mode, impGraymask, Show, chName, Invert):
# Attention!! impZone is absolutery 8bit!
    if Invert == False:
        zoneLTh = 0 # in Positive mask Luminosity mask zone low threshold must be 0!
    ip = impZone.getProcessor()
    mask = ByteProcessor(ip.width, ip.height)
    mask.setValue(255)
    if Show == "Yes":
        impZone.show()
# make Zone mask
    impZone.getProcessor().setThreshold(zoneLTh, zoneHTh, ImageProcessor.NO_LUT_UPDATE)
    print "Zone L & H", zoneLTh, zoneHTh
    roi = ThresholdToSelection.run(impZone)
    mask.setRoi(roi)
    mask.fill(roi)
    try:
        maskimp = ImagePlus("Mask", mask) 
#        only 8bit Zone mask made from ROI: Zone mask: maskimp
    except: # exception process for ROI missing
        IJ.error("Selected ROI (Region of Interest) is missing.")
        if Show == "Yes":
            impGraymask.hide()
            impZone.hide()
        impGraymask.close()
        impZone.close()
        return
    if Show == "Yes":
        maskimp.show() # show Zone mask

# deepen Zone Mask bit depth  (8bit to 16 or 32 bit)      
    if Mode == "16-bit":
        ImageConverter(maskimp).convertToGray16()
    elif Mode == "32-bit":
        ImageConverter(maskimp).convertToGray32()
    IJ.run(maskimp, "Multiply...", "value=" + str(convValue(1, Mode)))
#   after deepen bit depth, Save Zone Mask
    IJ.saveAsTiff(maskimp, folder + filename +"_" + chName + "_Zone" + str(int(zoneLTh))\
                                             + "-" + str(int(zoneHTh)) + "_Mask.tif")
    return maskimp

#----------------------------------------------
### RGB Manipulation 
def RGBmanipulation(imp0, folder,filename, baseCh, parm, colorCh, stackSize):
#  input Parameters
    if parm is not None: # if parm is not null, then get paraneters...
        Mode, YTh, blInvert, zoneLTh, zoneHTh, TrFac, Show = parm 
#     Mode: Image Format 16-bit or 8-bit
#     YTh: yellow(or blue) value threshold of mask
#     zoneLTh: lower threshold of mask
#     zoneHTh: upper threshold of mask
#     blinvert: make blue mask or not
#     Show: show working windows or not
# Windows Showing Mode
        imp0.show()
    impsCh = ChannelSplitter.split(imp0) #Channel Splitting of imp0
    for i in impsCh: # it needs to convert Grayscale image
        if Mode == "8-bit":
            ImageConverter(i).convertToGray8()
        elif Mode == "16-bit":
            ImageConverter(i).convertToGray16()
        else:
            ImageConverter(i).convertToGray32()
    if stackSize !=1: # if imp0 is not monochrome
        """ R ch.""" 
        impR =  impsCh[0]
        """ G ch.""" 
        impG =  impsCh[1]
        """ B ch. & B x 0.3""" 
        impB =  impsCh[2]
# Show Windows
#    imp0.close()
    if Show == "Yes":
        if stackSize !=1: # if imp0 is not monochrome
            impR.show()
            impG.show()
            impB.show()
        imp0.hide()
    if colorCh == "Blue":
        imp1 = Duplicator().run(impB)
        chName0 = "B"
    elif colorCh == "Red":
        imp1 = Duplicator().run(impR)
        chName0 = "R"
    elif colorCh == "Green":
        imp1 = Duplicator().run(impG)
        chName0 = "G"
    else: #Lightness or Luminosity
        imp1 = Duplicator().run(imp0)
        chName0 = "L"
        if colorCh == "Luminosity":
            chName0 = "Lm"
            IJ.run(imp1, "Conversions...", "weighted")
        if Mode == "16-bit":
            ImageConverter(imp1).convertToGray16()
        elif Mode == "8-bit":
            ImageConverter(imp1).convertToGray8()
        imp1.getProcessor().setLut(None)

#-----------------------------------
# Make Luminosity Mask
    impYmask, V = luminosity_mask_maker(imp1, zoneLTh, Mode, blInvert, Show, \
                                                                  TrFac, chName0)

# select base channel for deciding mask upper and lower threshold
    if baseCh == "Luminosity":
        impZone = Duplicator().run(imp0)
        IJ.run(impZone, "Conversions...", "weighted")
        chName1 ="Lm"
        ImageConverter(impZone).convertToRGB()
    elif baseCh == "Lightness":
        impZone = Duplicator().run(imp0)
        chName1 ="L"
        ImageConverter(impZone).convertToRGB()
    elif baseCh == "Blue":
        impZone = Duplicator().run(impB)
        chName1 ="B"
    elif baseCh == "Red":
        impZone = Duplicator().run(impR)
        chName1 ="R"
    elif baseCh == "Green":
        impZone = Duplicator().run(impG)
        chName1 ="G"
# if imp0 is not monochrome
    if stackSize != 1:
        impR.close()
        impG.close()
        impB.close()
    imp0.close()
#-----------------------------------------------
# impZone for Grayscale must be absolutely 8 bit!!
    ImageConverter(impZone).setDoScaling(True)
#           it is needs for 16bit image to down convert to 8bit
    ImageConverter(impZone).convertToGray8()
# call make_binary_mask
    maskimp = make_binary_mask(impZone, zoneLTh, zoneHTh, Mode, \
                                impYmask, Show, chName1, blInvert)
# maskimp may be 8bit or 16bit depending on original image bit
#--------------------------------------------------
# bind Zone mask and Yellow / blue mask
    ic = plugin.ImageCalculator()
    maskBindimp = ic.run("Min created", impYmask, maskimp) # bind masks
    if Show == "Yes":
        maskBindimp.show()
    IJ.saveAsTiff(maskBindimp, folder + filename +"_" + chName0 + "_Zone" + \
        str(int(zoneLTh)) + "-" + str(int(zoneHTh)) + "_" + chName1 + V + \
                                                          "_Mask_Binded.tif")
#     save binded mask
    if Show == "Yes":
        impYmask.hide()
        impZone.hide()
        maskimp.hide()
        maskBindimp.hide()
    impYmask.close()
    maskimp.close()
    maskBindimp.close()
    impZone.close()
#   End of def RGBmanipulation 
#----------------------------------------------
class LuminosityMaskPreviewer(AdjustmentListener, ItemListener, TextListener): 
         # previewer for binary mask and observe changes of dialog items
    def __init__(self, impPrv, impPrv1, slider_upper, slider_lower, C0, C1, \
        nfield_TrFac, checkbox_prv, checkbox_Bmask):
        """
             imp: an ImagePlus
             slider: a java.awt.Scrollbar UI element
             preview_checkbox: a java.awt.Checkbox controlling whether to
                                                 dynamically update the ImagePlus as the
                                                 scrollbar is updated, or not.
             nfield_TrFac: : a java.awt.TextField
        """

        self.impPrv = impPrv # main preview image / original image for binary mask
        self.impPrv1 = impPrv1 # original image for grayscale mask
        self.impPrv_binary = Duplicator().run(impPrv) # temporary binary mask image
        self.impPrv_Grayscalemask = Duplicator().run(impPrv1) 
                                                   # temporary grayscale mask image
        self.impPrv.setTitle("luminosity mask preview")
        self.impPrv_binary.setTitle("binary mask preview")
        self.impPrv_Grayscalemask.setTitle("grayscale mask preview")
        self.original_ip = impPrv.getProcessor().duplicate() 
                                                   # original image for binary mask
        self.slider_upper = slider_upper
        self.slider_lower = slider_lower
        self.C0 = C0 # positive mask color name
        self.C1 = C1 # negative mask color name
        self.C = C0  # active mask color name
        self.checkbox_prv = checkbox_prv # Preview checkbox
        self.checkbox_Bmask = checkbox_Bmask # Negative / Positive checkbox
        self.nfield_TrFac = nfield_TrFac # transparency factor
        textStrings2 = self.nfield_TrFac.getText() 
                             #  field for transparency factor for mask
        self.maskTrFac =  float(textStrings2) # transparency factor for mask

        self.original_ip = self.impPrv.getProcessor().duplicate() 
        self.imp_org = Duplicator().run(self.impPrv)
                                                    # store a copy


    def adjustmentValueChanged(self, event):
        """ event: an AdjustmentEvent with data on the state of the scroll bar. """
        preview = self.checkbox_prv.getState()
        if preview: # if preview available
            if (IJ.isMacOSX() == False) and event.getValueIsAdjusting():
                return 
                  # there are more scrollbar adjustment events queued already
            self.threshold() # make binary mask image

    def textValueChanged(self, event):
        """ event: a textValueChanged with data on the state of the numeric field. """
        try:
            textStrings2 = self.nfield_TrFac.getText()
            self.maskTrFac = float(textStrings2) # Transparency factor for mask
        except:
            return
        preview = self.checkbox_prv.getState()
        if preview:
            self.luminosity_prv() 
            # if preview is abailable and parameter is correct, make luminosity mask image 
        return

    def itemStateChanged(self, event):
        """ event: an ItemEvent with data on what happened to the checkbox. """
        Bmask = self.checkbox_Bmask.getState()
                # Bmask value: True/False
        if Bmask == True: # make negative mask
            self.C = self.C1
        else: # make positive mask
            self.C = self.C0
        self.luminosity_prv()  # make grayscale mask image
    
    def reset(self):
        """ Restore the original ImageProcessor """
        self.impPrv.setProcessor(self.original_ip)
    
    def threshold(self): # routine for making binary mask
        """ Execute the in-place making binary image of the ImagePlus. """
        ip = self.original_ip
        mask = ip.createProcessor(ip.width, ip.height)
        mask.BLACK_AND_WHITE_LUT
        mask.setValue(convValue(255, Mode))
        threshold_upper = self.slider_upper.getValue()
        threshold_lower = self.slider_lower.getValue()
        if threshold_upper < threshold_lower:
            threshold_upper = threshold_lower
        ip.setThreshold(threshold_lower, threshold_upper, \
            ImageProcessor.OVER_UNDER_LUT)
        roi = ThresholdToSelection().convert(ip) 
            # make ROI from threshold
        mask.setRoi(roi)
        mask.fill(roi)
        self.impPrv_binary.setProcessor(mask)
        self.impPrv_binary.getProcessor().setLut (None)
#     combine binary mask imega and grayscale mask image 
        ic =plugin.ImageCalculator()
        tmpPrv = ic.run("min created", self.impPrv_Grayscalemask, self.impPrv_binary)
        ip_tmpPrv = tmpPrv.getProcessor()
        self.impPrv.setProcessor(ip_tmpPrv) # set combined image for preview
#
        IJ.selectWindow(self.impPrv.getTitle())

    def luminosity_prv(self):
    # Make Grayscale Mask Image Files 
    # Target  Mask Threshold
        threshold_lower = self.slider_lower.getValue() # set lower threshold
        V = str(threshold_lower) # for threshold expression in file name
        impPrvTmp = Duplicator().run(self.impPrv1)
        IJ.run(impPrvTmp, "Subtract...","value=" + V)
        print "Luminosity self.C: " + self.C
        if self.C == self.C1: # for making inverted Negative mask
            IJ.run(impPrvTmp, "Divide...", "value=" + str(float(self.maskTrFac))) 
                # making mask more transparent (x 4 for Nega) < canceled
            impPrvTmp.getProcessor().invert() # make nega image
        else: # for making positive mask
            IJ.run(impPrvTmp, "Multiply...", "value=" + str(float(self.maskTrFac))) 
                # making mask more transparent (x 4) < canceled
        ip_PrvMask = impPrvTmp.getProcessor()
        ip_PrvMask.BLACK_AND_WHITE_LUT
        self.impPrv_Grayscalemask.setProcessor(ip_PrvMask)
#     combine grayscale image and binary image
        ic =plugin.ImageCalculator()
        #if self.impPrv_binary:
        impPrvTmp1 = ic.run("min created", impPrvTmp, self.impPrv_binary)
        ip_tmpPrv = impPrvTmp1.getProcessor()
        self.impPrv.setProcessor(ip_tmpPrv) # set combined image to preview

#
        self.impPrv.getProcessor().setLut (None)
        impPrvTmp.close()
        impPrvTmp1.close()
        IJ.selectWindow(self.impPrv.getTitle())

#---------------------------------------------
def thresholdImageUI(imp, imp1, Mode, C0, C1):
    """
      imp: preview image and original for binary mask
      imp1: original for grayscale mask
      C0: positive color name
      C1: negative color name
    """
    WM.getActiveWindow().toFront()
    gd = GenericDialog(C0 +" (" + C1 + ") Mask Maker / input threshold parameters")
    gd.addCheckbox("Making " + C1 + " Nega Mask", False)
    gd.addSlider("Upper Threshold of Mask", 0, 255, 128)
    gd.addSlider("Lower Threshold of Mask", 0, 255, 128)
    gd.addCheckbox("Preview", True)
    gd.addNumericField("Mask Transparency Factor", 1.0, 1)
    showWin = ["Yes","No"]
    gd.addChoice("Show working windows?", showWin, showWin[0])

    # The UI elements for the above two inputs
    slider_upper = gd.getSliders().get(0) # the 1st
    slider_lower = gd.getSliders().get(1) # the 2nd
    checkbox_Bmask = gd.getCheckboxes().get(0) #  chkbox for blue mask
    checkbox_prv = gd.getCheckboxes().get(1) # chkbox for preview
        #Attention! Slide Bar also has Numeric Field
    nfield_TrFac = gd.getNumericFields().get(2) # Transparency Factor
    choice_showWin = gd.getChoices().get(0)


    if not imp:
        print "Open an image first!"
        return


    #-- add Luminosity Mask Previewer--
    previewerL = LuminosityMaskPreviewer(imp, imp1, slider_upper, slider_lower,\
            C0, C1, nfield_TrFac, checkbox_prv, checkbox_Bmask)
    slider_upper.addAdjustmentListener(previewerL) # set listener for upper slider
    slider_lower.addAdjustmentListener(previewerL) # set listener for lower slider
    checkbox_Bmask.addItemListener(previewerL)
    checkbox_prv.addItemListener(previewerL)  # set listener for preview checkbox
    nfield_TrFac.addTextListener(previewerL)  
            # set listener for transparency fac. checkbox
#
    gd.showDialog()

    if gd.wasCanceled():
        previewerL.reset()
        Mode = "Stop"
        print "User canceled dialog!"
#    else:
#        previewer.threshold()
#    parm_Mode   = "Auto"
    YTh =  0
    blInvert = gd.getNextBoolean()
    zoneHTh =  gd.getNextNumber()
    zoneLTh =  gd.getNextNumber()
    TrFac =  gd.getNextNumber()
    Show =  gd.getNextChoice()
    imp.hide()
    imp.close()
    return Mode, YTh, blInvert, zoneLTh, zoneHTh, TrFac, str(Show)

#-------------------------------------------
class DialogPreviewer(ItemListener):    
    def __init__(self, choice1, choice2):    
        """ 
             choice: a java.awt.choice
        """    
        self.choice1 = choice1    
        self.choice2 = choice2    

 
    def itemStateChanged(self, event):    
        """ event: an ItemEvent with data on what happened to the checkbox. """    
        choice_item  = self.choice1.getSelectedItem()
        #print "Dialog Previewr: ", str(choice_item)
        if choice_item == "Luminosity":
            choice_item2 = "Luminosity"
        elif choice_item == "Blue":
            choice_item2 = "Blue"
        elif choice_item == "Green":
            choice_item2 = "Green"
        elif choice_item == "Red":
            choice_item2 = "Red"
        else: # Lightness
            choice_item2 = "Lightness"
        self.choice2.select(choice_item2)    
        return

#-----------------------------

def selectChannel(): # dialog of selecting target channel for mask upper & lower threshold
    gd = GenericDialog("Lightness Mask Maker / Target Channel Selection")
    scolorCh = ["Lightness", "Luminosity", "Red","Green","Blue"]
    gd.addChoice("Select Target Channel for Mask", scolorCh, scolorCh[0])
    sbaseCh = ["Lightness", "Luminosity","Red","Green","Blue"]
    gd.addChoice("Select Target Channel for deciding upper & lower threshold", sbaseCh, sbaseCh[0])

    # The UI elements for the above two inputs    
    choice1 = gd.getChoices().get(0)
    choice2 = gd.getChoices().get(1)
    
    previewer = DialogPreviewer(choice1, choice2)    
    choice1.addItemListener(previewer)    
 
    gd.showDialog()

    colorCh =  gd.getNextChoice()
    baseCh =  gd.getNextChoice()
    if gd.wasCanceled():
        colorCh =  "None"
        baseCh =  "None"

    return str(baseCh),  str(colorCh)

#----------main----------
window = WM.getActiveWindow()
Mode = "Auto"
if not window:
    od = OpenDialog("Choose a file")
    folder = od.getDirectory()
    filename = od.getFileName()
    if filename:
        path = folder + filename
        IJ.log(path)
        IJ.log(folder)
        IJ.log(filename)
# Always Open file with Bio-format Plugin
        imps = BF.openImagePlus(folder + filename) # Open file with Bio-format Plugin
        imp0 = imps[0] #imp0: Original Image 
        imp0.show()
    else:
        Mode = "Stop"
else: # target image is already open
    imp0 = IJ.getImage()
    filename = imp0.getTitle()
    try:
        folder = imp0.getOriginalFileInfo().directory
    except:
        folder = IJ.getDirectory("Output_directory")
if Mode == "Auto":
    bitdepth = imp0.getBitDepth()
    if bitdepth == 8:
            Mode = "8-bit"
    elif bitdepth ==16:
            Mode = "16-bit"
    elif bitdepth ==32:
            Mode = "32-bit"
    else:
            Mode = "16-bit"
stackSize = imp0.getStackSize() # imp0 is color or monochrome?
if Mode <> "Stop": # if user did not cancel execute process
    impprv = Duplicator().run(imp0) # impprv: preview imp
    scale = 1200.0 / impprv.getWidth() 
                          # preview image scaling value (width = 1200px fixed)
    new_height = impprv.getHeight() * scale # preview image height
    new_height = int(new_height)
    tmpimpprv = impprv.resize(1200, new_height, "none")
    ImageConverter(tmpimpprv).convertToRGB() # convert to 8bit RGB
    impprvsCh = ChannelSplitter.split(tmpimpprv)
    
    if stackSize == 1: #imp0 is monochrome
        baseCh = "Luminosity"
        colorCh = "Luminosity"
    else:  #imp0 is not monochrome
        baseCh, colorCh = selectChannel()
        """ select target channel for deciding upper & lower threshold """
    if colorCh == "Blue":
        C0 = "Blue"
        C1 = "Blue inverted"
        impPrv1 = Duplicator().run(impprvsCh[2])
    elif colorCh == "Red":
        C0 = "Red"
        C1 = "Red inverted"
        impPrv1 = Duplicator().run(impprvsCh[0])
    elif colorCh == "Green":
        C0 = "Green"
        C1 = "Green inverted"
        impPrv1 = Duplicator().run(impprvsCh[1])
    elif colorCh == "Luminosity":
        C0 = "Luminosity Positive"
        C1 = "Luminosity Negative"
        impPrv1 = Duplicator().run(tmpimpprv)
        if stackSize !=1:
            IJ.run(impPrv1,"Conversions...","weighted")
    elif colorCh == "Lightness":
        C0 = "Lightness Positive"
        C1 = "Lightness Negative"
        impPrv1 = Duplicator().run(tmpimpprv)
    ImageConverter(impPrv1).convertToGray8()

    if baseCh != "None": # if process is not canceled
        imp0.hide()
        if baseCh == "Blue":
            impPrv0 = Duplicator().run(impprvsCh[2])
        elif baseCh == "Red":
            impPrv0 = Duplicator().run(impprvsCh[0])
        elif baseCh == "Green":
            impPrv0 = Duplicator().run(impprvsCh[1])
        else:
            impPrv0 = Duplicator().run(tmpimpprv)
        ImageConverter(impPrv0).convertToGray8()
        impPrv0.show()
        parm = thresholdImageUI(impPrv0, impPrv1, Mode, C0, C1)
    else: # if process is  canceled
        parm = []
        parm.append("Stop")
        imp0.hide()
        imp0.close()
    impPrv0.hide()
    impPrv0.close()
    
    if parm[0] <> "Stop":
        RGBmanipulation(imp0, folder, filename, baseCh, parm, colorCh, stackSize)
else:
    IJ.log("User canceled")
IJ.log("Finish")

